#include "StdAfx.h"
#include "syslog.h"

#include "gpu_i.h"

void cmdSTP(unsigned char * baseAddr)
{
	unsigned long gdata = ((unsigned long*)baseAddr)[0];
	syslog(77," CMD: STP: %x",gdata&0xffffff);
}

struct TransparencySettings{
	int src;
	int dst;
	byte alpha;
	BOOL change;
};
TransparencySettings TransRate[4]={
	{GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA,128,FALSE},
	{GL_ONE,GL_ONE_MINUS_SRC_ALPHA,0,FALSE},
	{GL_ZERO,GL_ONE_MINUS_SRC_COLOR,0,TRUE},
	{GL_SRC_ALPHA,GL_ONE,64,TRUE}
};
BOOL transparent=FALSE;
byte gAlpha;

void cmdTexturePage(unsigned char * baseAddr)
{
	unsigned long gdata = ((unsigned long*)baseAddr)[0];
	textAddrX = (gdata&0xf)<<6;
	// this one is only for speeding up other calculations
	textAddrX2 = (gdata&0xf)<<7;
	textAddrY = (((gdata)<<4)&0x100);//+(((gdata)>>2)&0x200);
	textTP = (gdata & 0x180) >> 7;
	long tempABR;
	tempABR = (gdata >> 5) & 0x3;
	if(textABR!=tempABR){
		textABR=tempABR;
		glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
		gAlpha=TransRate[textABR].alpha;
		transparent=TRUE;
	}
	/*switch(textABR){
	  case 0:
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		gAlpha=128;
		break;
	  case 1:
		glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA);
		gAlpha=0;
		break;
	  case 2:
		glBlendFunc(GL_ZERO,GL_ONE_MINUS_SRC_COLOR);
		gAlpha=0;
		break;
	  case 3:
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
		gAlpha=64;
		break;
	}*/
	textREST = (gdata&0x00ffffff)>>9;
	//syslog(77," CMD: TPage: %d,%d TP=%d ABR=%d REST=%x\n",textAddrX,textAddrY,textTP,textABR,(gdata&0x00ffffff)>>9);
	return;
}

void cmdTextureWindow(unsigned char *baseAddr)
{
	unsigned long gdata = ((unsigned long*)baseAddr)[0];
	//syslog(77,"CDM: TWin: %x *NI*", gdata&0xffffff);
}

void cmdDrawAreaStart(unsigned char * baseAddr)
{
	unsigned long gdata = ((unsigned long*)baseAddr)[0];
	drawX = gdata & 0x3ff;
	drawY = (gdata>>10)&0x1ff;
	dispOffset=-(short)drawY;

	//syslog(77," CMD: DrawArea Start %d,%d %x",drawX,drawY,gdata);
}

void cmdDrawAreaEnd(unsigned char * baseAddr)
{
	unsigned long gdata = ((unsigned long*)baseAddr)[0];
	drawW = gdata & 0x3ff;
	drawH = (gdata>>10)&0x1ff;

	if(isFirst==FALSE){
		GLdouble equation[4]={0.0,0.0,0.0,0.0};
		equation[0]=-1.0;
		equation[3]=(GLdouble)(drawW-drawX+1);
		glClipPlane(GL_CLIP_PLANE0,equation);
		equation[0]=0.0;
		equation[1]=-1.0;
		equation[3]=(GLdouble)(drawH-drawY+1);
		glClipPlane(GL_CLIP_PLANE1,equation);
	}

	//syslog(77," CMD: DrawArea End %d,%d",drawW,drawH);
}

void cmdDrawOffset(unsigned char * baseAddr)
{
	unsigned long gdata = ((unsigned long*)baseAddr)[0];
	ofsX = (short)(gdata & 0x3ff);
	ofsY = (short)((gdata>>11)&0x1ff);
	//syslog(77," CMD: Draw Offset %d,%d",ofsX,ofsY,gdata);
}

void primLoadImage(unsigned char * baseAddr)
{
	unsigned long *gpuData = ((unsigned long *) baseAddr);
	imageX0 = (short)(gpuData[1] & 0xffff);
	imageY0 = (short)(gpuData[1]>>16) & 0xffff;
	imageX1 = (short)(gpuData[2] & 0xffff);
	imageY1 = (short)((gpuData[2]>>16) & 0xffff);

	for(int i=0;i<gpuConfig.nMaxTextures;i++){
		if(((texture[i].textAddrX+255)>=imageX0)
		  &&((texture[i].textAddrY+255)>=imageY0)
		  &&(texture[i].textAddrX<=(imageX0+imageX1))
		  &&(texture[i].textAddrY<=(imageY0+imageY1))){
			texture[i].Update=TRUE;
		}
	}

	//syslog(1," CMD: MEM->VRAM %d,%d %d,%d\n",imageX0,imageY0,imageX1,imageY1);
	imTX=imageX0; imTY=imageY0;
	imTXc=imageX1; imTYc=imageY1;
	imageTransfer = 1;
}

void primStoreImage(unsigned char * baseAddr)
{
	unsigned long *gpuData = ((unsigned long *) baseAddr);

	imageX0 = (short)(gpuData[1] & 0x3ff);
	imageY0 = (short)(gpuData[1]>>16) & 0x1ff;
	imageX1 = (short)(gpuData[2] & 0xffff);
	imageY1 = (short)((gpuData[2]>>16) & 0xffff);

	//syslog(1," CMD: VRAM->MEM %d,%d %d,%d\n",imageX0,imageY0,imageX1,imageY1);
	imTX=imageX0; imTY=imageY0;
	imTXc=imageX1; imTYc=imageY1;
	imSize=imageY1*imageX1/2;

	imageTransfer = 2;
	GPUstatusRet|=0x08000000;
}

void primMoveImage(unsigned char * baseAddr)
{
    // 0x80 image move (VRAM->VRAM)
	unsigned long *gpuData = ((unsigned long *) baseAddr);

	unsigned short imageY0,imageX0,imageY1,imageX1,imageSX,imageSY,i,j;

    imageX0 = (short)(gpuData[1] & 0x3ff);
    imageY0 = (short)(gpuData[1]>>16) & 0x1ff;

    imageX1 = (short)(gpuData[2] & 0x3ff);
    imageY1 = (short)((gpuData[2]>>16) & 0x1ff);

    imageSY = (short)((gpuData[3]>>16) & 0xffff);
    imageSX = (short)(gpuData[3] & 0xffff);

	for(i=0;i<gpuConfig.nMaxTextures;i++){
		if(((texture[i].textAddrX+255)>=imageX1)
		  &&((texture[i].textAddrY+255)>=imageY1)
		  &&(texture[i].textAddrX<=(imageX1+imageSX))
		  &&(texture[i].textAddrY<=(imageY1+imageSY))){
			texture[i].Update=TRUE;
		}
	}

	//syslog(1," CMD: VRAM->VRAM %d,%d -> %d,%d (%d,%d)\n",imageX0,imageY0,imageX1,imageY1,imageSX,imageSY);
    
    if((imageY0+imageSY)>512) imageSY=512-imageY0;
    if((imageY1+imageSY)>512) imageSY=512-imageY1;
    if((imageX0+imageSX)>1024) imageSX=1024-imageX0;
    if((imageX1+imageSX)>1024) imageSX=1024-imageX1;
    for(j=0;j<imageSY;j++)
		for(i=0;i<imageSX;i++)
			psxVuw[(imageY1+j)<<10+imageX1+i]=psxVuw[(imageY0+j)<<10+imageX0+i];
}




void primTileS(unsigned char * baseAddr)
{
	syslog(2,"Tile S *NI*");//  : %04x,%04x - %04x,%04x",baseAddrW[addr+2],baseAddrW[addr+3],baseAddrW[addr+4],baseAddrW[addr+5]);
	//syslog(2,"Tile S *NI* : %04x,%04x - %04x,%04x",baseAddrW[addr+2],baseAddrW[addr+3],baseAddrW[addr+4],baseAddrW[addr+5]);
	return;
}

void primBlkFill(unsigned char * baseAddr)
{
	unsigned short * gpuPoint = (unsigned short*) baseAddr;
	unsigned long *gpuData = ((unsigned long *) baseAddr);

	//syslog(2,"BlkFill : %04x,%04x - %04x,%04x",gpuPoint[2],gpuPoint[3],gpuPoint[4],gpuPoint[5]);

/*
	this function is something like clear_with_given_color
	and according to baseAddr
	struct packet
	{
		unsigned char R,G,B,code;
		unsigned short X,Y,W,H;
	}
	this function is not clipped bby drawing area, accesses whole VRAM
*/
	glColor3ub(baseAddr[0],baseAddr[1],baseAddr[2]);
	glBegin(GL_POLYGON);
	  glVertex2s(gpuPoint[2],gpuPoint[3]+dispOffset);
	  glVertex2s(gpuPoint[2]+gpuPoint[4]-1,gpuPoint[3]+dispOffset);
	  glVertex2s(gpuPoint[2]+gpuPoint[4]-1,gpuPoint[3]+gpuPoint[5]-1+dispOffset);
	  glVertex2s(gpuPoint[2],gpuPoint[3]+gpuPoint[5]-1+dispOffset);
	glEnd();
}

unsigned long torgba[65536];
unsigned long image[65536];
char curText=0;
texturecache texture[32];
GLuint nullid;

void setupTexture(long clutP){
	int i;
	long sprCY,sprCX,tbw;
	short color,tC;
	int ctext;
	uPointers psxVOffset,psxVOffset2;

	if(gpuConfig.nMaxTextures==0)return;
	if((textTP&2)==2)clutP=nullclutP;
	ctext=curText;
	for(i=0;i<gpuConfig.nMaxTextures;i++){
		if(texture[i].textAddrX==textAddrX
		  &&texture[i].textAddrY==textAddrY){
			if(texture[i].Update==TRUE||(gpuConfig.bCheckPalette&&clutP!=texture[i].clutP)||textTP!=texture[i].textTP){
				ctext=i;
				break;
			}
			glBindTexture(GL_TEXTURE_2D,texture[i].id);
			return;
		}
	}

	texture[ctext].textAddrX=textAddrX;
	texture[ctext].textAddrY=textAddrY;
	texture[ctext].textTP=textTP;
	texture[ctext].clutP=clutP;
	texture[ctext].Update=FALSE;
	glBindTexture(GL_TEXTURE_2D,texture[ctext].id);
	if(ctext==curText){
		curText++;
		if(curText>=gpuConfig.nMaxTextures)curText=0;
	}
	tbw=256;
	switch(textTP){
	  case 1:
		if(textAddrX==960)tbw=128;
		break;
	  case 2:
		if(textAddrX==960)tbw=64;
		else if(textAddrX==896)tbw=128;
		else if(textAddrX==832)tbw=192;
		break;
	}
	psxVOffset.w=psxVuw+(textAddrY<<10)+textAddrX;
	if(gpuConfig.bTextureHalf){
		switch(textTP){
		  case 0:
			for (sprCY=0;sprCY<128;sprCY++){
				psxVOffset2.b=psxVOffset.b;
				for (sprCX=0;sprCX<128;sprCX++){
					tC = (*psxVOffset2.b)&0x0f;
					image[(sprCY<<7)+sprCX] = torgba[psxVuw[clutP+tC]];
					psxVOffset2.b++;
				}
				psxVOffset.b+=(2048*2);
			}
			break;
		  case 1:
			for (sprCY=0;sprCY<128;sprCY++){
				psxVOffset2.b=psxVOffset.b;
				for (sprCX=0;sprCX<tbw;sprCX+=2){ 
					tC = *psxVOffset2.b;
					image[(sprCY<<7)+(sprCX>>1)] = torgba[psxVuw[clutP+tC]];
					psxVOffset2.b+=2;
				}
				psxVOffset.b+=(2048*2);
			}
			break;
		  case 2:
			for (sprCY=0;sprCY<128;sprCY++){
				psxVOffset2.w=psxVOffset.w;
				for (sprCX=0;sprCX<tbw;sprCX+=2){
					color = *psxVOffset2.w;
					image[(sprCY<<7)+(sprCX>>1)] = torgba[color];
					psxVOffset2.w+=2;
				}
				psxVOffset.w+=(1024*2);
			}
			break;
		}
		glTexSubImage2D(GL_TEXTURE_2D,0,0,0,128,128,GL_RGBA,GL_UNSIGNED_BYTE,image);
	}else{
		switch(textTP){
		  case 0:
			for (sprCY=0;sprCY<256;sprCY++){
				psxVOffset2.b=psxVOffset.b;
				for (sprCX=0;sprCX<128;sprCX++){
					tC = (*psxVOffset2.b)&0x0f;
					image[(sprCY<<8)+(sprCX<<1)] = torgba[psxVuw[clutP+tC]];
					tC = (*psxVOffset2.b>>4)&0x0f;
					image[(sprCY<<8)+(sprCX<<1)+1] = torgba[psxVuw[clutP+tC]];
					psxVOffset2.b++;
				}
				psxVOffset.b+=2048;
			}
			break;
		  case 1:
			for (sprCY=0;sprCY<256;sprCY++){
				psxVOffset2.b=psxVOffset.b;
				for (sprCX=0;sprCX<tbw;sprCX++){
					tC = *psxVOffset2.b;
					image[(sprCY<<8)+sprCX] = torgba[psxVuw[clutP+tC]];
					psxVOffset2.b++;
				}
				psxVOffset.b+=2048;
			}
			break;
		  case 2:
			for (sprCY=0;sprCY<256;sprCY++){
				psxVOffset2.w=psxVOffset.w;
				for (sprCX=0;sprCX<tbw;sprCX++){
					color = *psxVOffset2.w;
					image[(sprCY<<8)+sprCX] = torgba[color];
					psxVOffset2.w++;
				}
				psxVOffset.w+=1024;
			}
			break;
		}
		glTexSubImage2D(GL_TEXTURE_2D,0,0,0,256,256,GL_RGBA,GL_UNSIGNED_BYTE,image);
	}
}

void primSprt8(unsigned char * baseAddr)
{
/* 
	fixed size (8x8) sprite
*/
	unsigned long *gpuData = (unsigned long *)baseAddr;
	unsigned short *gpuPoint = (unsigned short*) baseAddr;
	long clutP;
	short tx,ty;

	//syslog(2,"Sprt8x8 : %d,%d %02x",gpuPoint[2],gpuPoint[3],baseAddr[3]);
    clutP  = (gpuData[2]>>12) & 0x7fff0;//0xffff0;
    tx = (short)(gpuData[2] & 0x000000ff);
    ty = (short)((gpuData[2]>>8) & 0x000000ff);

	setupTexture(clutP);
	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		if(baseAddr[3]&1){
			glColor4ub(255,255,255,gAlpha);
		}else{
			glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],gAlpha);
		}
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		if(baseAddr[3]&1){
			glColor3ub(255,255,255);
		}else{
			glColor3ub(baseAddr[0],baseAddr[1],baseAddr[2]);
		}
	}
	/*switch(baseAddr[3]&3){
	  case 0:
		glColor3ub(baseAddr[0],baseAddr[1],baseAddr[2]);
		break;
	  case 1:
		glColor3ub(255,255,255);
		break;
	  case 2:
		glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],gAlpha);
		break;
	  case 3:
		glColor4ub(255,255,255,gAlpha);
		break;
	}*/
	glBegin(GL_POLYGON);
	  glTexCoord2s(tx,ty);
	  glVertex2s(gpuPoint[2],gpuPoint[3]+dispOffset+ofsY);
	  glTexCoord2s(tx+8,ty);
	  glVertex2s(gpuPoint[2]+8,gpuPoint[3]+dispOffset+ofsY);
	  glTexCoord2s(tx+8,ty+8);
	  glVertex2s(gpuPoint[2]+8,gpuPoint[3]+8+dispOffset+ofsY);
	  glTexCoord2s(tx,ty+8);
	  glVertex2s(gpuPoint[2],gpuPoint[3]+8+dispOffset+ofsY);
	glEnd();
	glBindTexture(GL_TEXTURE_2D,nullid);
}

void primSprt16(unsigned char * baseAddr)
{
/* 
	fixed size (16x16) sprite
*/
	unsigned long *gpuData = (unsigned long *)baseAddr;
	unsigned short *gpuPoint = (unsigned short*) baseAddr;
	long clutP;
	short tx,ty;

	//syslog(2,"Sprt16x16 : %d,%d %02x",gpuPoint[2],gpuPoint[3],baseAddr[3]);
    clutP  = (gpuData[2]>>12) & 0x7fff0;//0xffff0;
    tx = (short)(gpuData[2] & 0x000000ff);
    ty = (short)((gpuData[2]>>8) & 0x000000ff);

	setupTexture(clutP);
	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		if(baseAddr[3]&1){
			glColor4ub(255,255,255,gAlpha);
		}else{
			glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],gAlpha);
		}
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		if(baseAddr[3]&1){
			glColor3ub(255,255,255);
		}else{
			glColor3ub(baseAddr[0],baseAddr[1],baseAddr[2]);
		}
	}
	glBegin(GL_POLYGON);
	  glTexCoord2s(tx,ty);
	  glVertex2s(gpuPoint[2],gpuPoint[3]+dispOffset+ofsY);
	  glTexCoord2s(tx+16,ty);
	  glVertex2s(gpuPoint[2]+16,gpuPoint[3]+dispOffset+ofsY);
	  glTexCoord2s(tx+16,ty+16);
	  glVertex2s(gpuPoint[2]+16,gpuPoint[3]+16+dispOffset+ofsY);
	  glTexCoord2s(tx,ty+16);
	  glVertex2s(gpuPoint[2],gpuPoint[3]+16+dispOffset+ofsY);
	glEnd();
	glBindTexture(GL_TEXTURE_2D,nullid);
}

void primSprtS(unsigned char * baseAddr)
{
/* 
	free size (up to 256x256) sprite
*/
	unsigned long *gpuData = (unsigned long *)baseAddr;
	unsigned short *gpuPoint = (unsigned short*) baseAddr;
	long clutP;
	short tx,ty,sprtW,sprtH;

    clutP  = (gpuData[2]>>12) & 0x7fff0;//0xffff0;
    tx = (short)(gpuData[2] & 0x000000ff);
    ty = (short)((gpuData[2]>>8) & 0x000000ff);
	sprtW = (short)(gpuData[3] & 0x3ff);
	sprtH = (short)((gpuData[3]>>16) & 0x1ff);

	//syslog(2,"Sprt : %04x,%04x - %04x,%04x code=%02x pal=%05x",gpuPoint[2],gpuPoint[3],gpuPoint[6],gpuPoint[7],baseAddr[3],clutP);
	/*if(textAddrX==320&&textAddrY==0){
		syslog(2,"Sprt : %d,%d - %d,%d pal=%05x",gpuPoint[2],gpuPoint[3],gpuPoint[6],gpuPoint[7],gpuData[2]>>12);
	}*/

	setupTexture(clutP);
	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		if(baseAddr[3]&1){
			glColor4ub(255,255,255,gAlpha);
		}else{
			glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],gAlpha);
		}
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		if(baseAddr[3]&1){
			glColor3ub(255,255,255);
		}else{
			glColor3ub(baseAddr[0],baseAddr[1],baseAddr[2]);
		}
	}
	glBegin(GL_POLYGON);
	  glTexCoord2s(tx,ty);
	  glVertex2s(gpuPoint[2]+ofsX,gpuPoint[3]+dispOffset+ofsY);
	  glTexCoord2s(tx+sprtW,ty);
	  glVertex2s(gpuPoint[2]+sprtW+ofsX,gpuPoint[3]+dispOffset+ofsY);
	  glTexCoord2s(tx+sprtW,ty+sprtH);
	  glVertex2s(gpuPoint[2]+sprtW+ofsX,gpuPoint[3]+sprtH+dispOffset+ofsY);
	  glTexCoord2s(tx,ty+sprtH);
	  glVertex2s(gpuPoint[2]+ofsX,gpuPoint[3]+sprtH+dispOffset+ofsY);
	glEnd();
	glBindTexture(GL_TEXTURE_2D,nullid);
}


void primLineF2(unsigned char *baseAddr)
{
/* 
	Flat Line 2 vertices
*/
	short *gpuPoint=((short *) baseAddr);
	short ly0,ly1,lx0,lx1;

	lx0 = gpuPoint[2]+ofsX;
	ly0 = gpuPoint[3]+ofsY;
	lx1 = gpuPoint[4]+ofsX;
	ly1 = gpuPoint[5]+ofsY;

	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],gAlpha);
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		glColor3ub(baseAddr[0],baseAddr[1],baseAddr[2]);
	}
	/*if(baseAddr[3]&2){
		glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],gAlpha);
	}else{
		glColor3ub(baseAddr[0],baseAddr[1],baseAddr[2]);
	}*/
	glBegin(GL_LINES);
		glVertex2s(lx0,ly0+dispOffset);
		glVertex2s(lx1,ly1+dispOffset);
	glEnd();
}

void primLineG2(unsigned char *baseAddr)
{
/* 
	Goraud Line 2 vertices
*/
	short *gpuPoint=((short *) baseAddr);
	short ly0,ly1,lx0,lx1;
	unsigned char alpha;

	lx0 = gpuPoint[2]+ofsX;
	ly0 = gpuPoint[3]+ofsY;
	lx1 = gpuPoint[6]+ofsX;
	ly1 = gpuPoint[7]+ofsY;

	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		alpha=gAlpha;
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		alpha=255;
	}
	/*if(baseAddr[3]&2){
		alpha=gAlpha;
	}else{
		alpha=255;
	}*/
	glBegin(GL_LINES);
		glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],alpha);
		glVertex2s(lx0,ly0+dispOffset);
		glColor4ub(baseAddr[8],baseAddr[9],baseAddr[10],alpha);
		glVertex2s(lx1,ly1+dispOffset);
	glEnd();
}

void primPolyF3(unsigned char *baseAddr)
{	
/* 
	Flat Shaded Polygon 3 vertices
*/
	short *gpuPoint=((short *) baseAddr);
	short ly0,ly1,lx0,lx1,ly2,lx2;

	lx0 = gpuPoint[2]+ofsX;
	ly0 = gpuPoint[3]+ofsY;
	lx1 = gpuPoint[4]+ofsX;
	ly1 = gpuPoint[5]+ofsY;
	lx2 = gpuPoint[6]+ofsX;
	ly2 = gpuPoint[7]+ofsY;
	//syslog(77,"!GPU! P3 F %d,%d %d,%d %d,%d\n",lx0,ly0,lx1,ly1,lx2,ly2);

	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],gAlpha);
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		glColor3ub(baseAddr[0],baseAddr[1],baseAddr[2]);
	}
	glBegin(GL_POLYGON);
		glVertex2s(lx0,ly0+dispOffset);
		glVertex2s(lx1,ly1+dispOffset);
		glVertex2s(lx2,ly2+dispOffset);
	glEnd();
}

void primPolyF4(unsigned char *baseAddr)
{
/* 
	Flat Shaded Polygon 4 vertices
*/
	short *gpuPoint=((short *) baseAddr);
	short ly0,ly1,ly2,ly3,lx0,lx1,lx2,lx3;

	lx0 = gpuPoint[2]+ofsX;
	ly0 = gpuPoint[3]+ofsY;
	lx1 = gpuPoint[4]+ofsX;
	ly1 = gpuPoint[5]+ofsY;
	lx2 = gpuPoint[6]+ofsX;
	ly2 = gpuPoint[7]+ofsY;
	lx3 = gpuPoint[8]+ofsX;
	ly3 = gpuPoint[9]+ofsY;
	//syslog(77,"!GPU! P4 F %d,%d %d,%d %d,%d %d,%d %6x\n",lx0,ly0,lx1,ly1,lx2,ly2,lx3,ly3,gpuData[0]);

	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],gAlpha);
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		glColor3ub(baseAddr[0],baseAddr[1],baseAddr[2]);
	}
	glBegin(GL_POLYGON);
		glVertex2s(lx0,ly0+dispOffset);
		glVertex2s(lx1,ly1+dispOffset);
		glVertex2s(lx3,ly3+dispOffset);
		glVertex2s(lx2,ly2+dispOffset);
	glEnd();
}

void primPolyG3(unsigned char *baseAddr)
{	
/* 
	Goraud Shaded Polygon 3 vertices
*/
	short *gpuPoint=((short *) baseAddr);
	short ly0,ly1,lx0,lx1,ly2,lx2;
	unsigned char alpha;

	lx0 = gpuPoint[2]+ofsX;
	ly0 = gpuPoint[3]+ofsY;
	lx1 = gpuPoint[6]+ofsX;
	ly1 = gpuPoint[7]+ofsY;
	lx2 = gpuPoint[10]+ofsX;
	ly2 = gpuPoint[11]+ofsY;
	//syslog(77,"!GPU! P3 G %d,%d %d,%d %d,%d\n",lx0,ly0,lx1,ly1,lx2,ly2);

	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		alpha=gAlpha;
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		alpha=255;
	}
	glBegin(GL_POLYGON);
		glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],alpha);
		glVertex2s(lx0,ly0+dispOffset);
		glColor4ub(baseAddr[8],baseAddr[9],baseAddr[10],alpha);
		glVertex2s(lx1,ly1+dispOffset);
		glColor4ub(baseAddr[16],baseAddr[17],baseAddr[18],alpha);
		glVertex2s(lx2,ly2+dispOffset);
	glEnd();
}

void primPolyG4(unsigned char * baseAddr)
{
/* 
	Goraud Shaded Polygon 4 vertices
*/
	short *gpuPoint=((short *) baseAddr);
	short ly0,ly1,ly2,ly3,lx0,lx1,lx2,lx3;
	unsigned char alpha;

	lx0 = gpuPoint[2]+ofsX;
	ly0 = gpuPoint[3]+ofsY;
	lx1 = gpuPoint[6]+ofsX;
	ly1 = gpuPoint[7]+ofsY;
	lx2 = gpuPoint[10]+ofsX;
	ly2 = gpuPoint[11]+ofsY;
	lx3 = gpuPoint[14]+ofsX;
	ly3 = gpuPoint[15]+ofsY;

	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		alpha=gAlpha;
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		alpha=255;
	}
	glBegin(GL_POLYGON);
		glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],alpha);
		glVertex2s(lx0,ly0+dispOffset);
		glColor4ub(baseAddr[8],baseAddr[9],baseAddr[10],alpha);
		glVertex2s(lx1,ly1+dispOffset);
		glColor4ub(baseAddr[24],baseAddr[25],baseAddr[26],alpha);
		glVertex2s(lx3,ly3+dispOffset);
		glColor4ub(baseAddr[16],baseAddr[17],baseAddr[18],alpha);
		glVertex2s(lx2,ly2+dispOffset);
	glEnd();
}

void primPolyFT3(unsigned char * baseAddr)
{
/* 
	Flat Shaded Textured Polygon 3 vertices
*/
	unsigned long *gpuData = (unsigned long *)baseAddr;
	short *gpuPoint=((short *) baseAddr);
	long clutP;
	short lx0,ly0,lx1,ly1,lx2,ly2;
	short tx0,ty0,tx1,ty1,tx2,ty2;
	unsigned short gpuDataX;

	lx0 = gpuPoint[2]+ofsX;
	ly0 = gpuPoint[3]+ofsY;
	lx1 = gpuPoint[6]+ofsX;
	ly1 = gpuPoint[7]+ofsY;
	lx2 = gpuPoint[10]+ofsX;
	ly2 = gpuPoint[11]+ofsY;

	tx0 = (short)(gpuData[2] & 0xff);
	ty0 = (short)((gpuData[2]>>8) & 0xff);
	tx1 = (short)(gpuData[4] & 0xff);
	ty1 = (short)((gpuData[4]>>8) & 0xff);
	tx2 = (short)(gpuData[6] & 0xff);
	ty2 = (short)((gpuData[6]>>8) & 0xff);
	//syslog(77,"!GPU! P3 F %d,%d %d,%d %d,%d\n",lx0,ly0,lx1,ly1,lx2,ly2);
	gpuDataX=(unsigned short)(gpuData[4]>>16);
	textAddrX = (gpuDataX&0xf)<<6;
	textAddrX2 = (gpuDataX&0xf)<<7;
	textAddrY = (((gpuDataX)<<4)&0x100);//+(((gpuDataX)>>2)&0x200);
	textTP = (gpuDataX & 0x180) >> 7;
	long tempABR;
	tempABR = (gpuDataX >>5) & 0x3;
	if(textABR!=tempABR){
		textABR=tempABR;
		glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
		gAlpha=TransRate[textABR].alpha;
		transparent=TRUE;
	}
	/*switch(textABR){
	  case 0:
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		gAlpha=128;
		break;
	  case 1:
		glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA);
		gAlpha=0;
		break;
	  case 2:
		glBlendFunc(GL_ZERO,GL_ONE_MINUS_SRC_COLOR);
		gAlpha=0;
		break;
	  case 3:
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
		gAlpha=64;
		break;
	}*/
	//syslog(77," CMD: P3TF: %d,%d TP=%d",textAddrX,textAddrY,textTP);
	//syslog(77," CMD: P3TF: %d,%d %02x",lx0,ly0,baseAddr[3]);

    clutP  = (gpuData[2]>>12) & 0x7fff0;//0xffff0;
	setupTexture(clutP);
	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		if(baseAddr[3]&1){
			glColor4ub(255,255,255,gAlpha);
		}else{
			glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],gAlpha);
		}
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		if(baseAddr[3]&1){
			glColor3ub(255,255,255);
		}else{
			glColor3ub(baseAddr[0],baseAddr[1],baseAddr[2]);
		}
	}
	glBegin(GL_POLYGON);
		glTexCoord2s(tx0,ty0);
		glVertex2s(lx0,ly0+dispOffset);
		glTexCoord2s(tx1,ty1);
		glVertex2s(lx1,ly1+dispOffset);
		glTexCoord2s(tx2,ty2);
		glVertex2s(lx2,ly2+dispOffset);
	glEnd();
	glBindTexture(GL_TEXTURE_2D,nullid);
}


void primPolyFT4(unsigned char * baseAddr)
{
/* 
	Flat Shaded Textured Polygon 4 vertices
*/
	unsigned long *gpuData = (unsigned long *)baseAddr;
	short *gpuPoint=((short *) baseAddr);
	long clutP;
	short lx0,ly0,lx1,ly1,lx2,ly2,lx3,ly3;
	short tx0,ty0,tx1,ty1,tx2,ty2,tx3,ty3;
	unsigned short gpuDataX;

	lx0 = gpuPoint[2]+ofsX;
	ly0 = gpuPoint[3]+ofsY;
	lx1 = gpuPoint[6]+ofsX;
	ly1 = gpuPoint[7]+ofsY;
	lx2 = gpuPoint[10]+ofsX;
	ly2 = gpuPoint[11]+ofsY;
	lx3 = gpuPoint[14]+ofsX;
	ly3 = gpuPoint[15]+ofsY;

	tx0 = (short)(gpuData[2] & 0xff);
	ty0 = (short)((gpuData[2]>>8) & 0xff);
	tx1 = (short)(gpuData[4] & 0xff);
	ty1 = (short)((gpuData[4]>>8) & 0xff);
	tx2 = (short)(gpuData[6] & 0xff);
	ty2 = (short)((gpuData[6]>>8) & 0xff);
	tx3 = (short)(gpuData[8] & 0xff);
	ty3 = (short)((gpuData[8]>>8) & 0xff);
	//syslog(77,"!GPU! P4 F %d,%d %d,%d %d,%d %d,%d %6x\n",tx0,ty0,tx1,ty1,tx2,ty2,tx3,ty3,gpuData[0]);

	gpuDataX=(unsigned short)(gpuData[4]>>16);
	textAddrX = (gpuDataX&0xf)<<6;
	textAddrX2 = (gpuDataX&0xf)<<7;
	textAddrY = (((gpuDataX)<<4)&0x100);//+(((gpuDataX)>>2)&0x200);
	textTP = (gpuDataX & 0x180) >> 7;
	long tempABR;
	tempABR = (gpuDataX >> 5) & 0x3;
	if(textABR!=tempABR){
		textABR=tempABR;
		glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
		gAlpha=TransRate[textABR].alpha;
		transparent=TRUE;
	}
	/*switch(textABR){
	  case 0:
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		gAlpha=128;
		break;
	  case 1:
		glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA);
		gAlpha=0;
		break;
	  case 2:
		glBlendFunc(GL_ZERO,GL_ONE_MINUS_SRC_COLOR);
		gAlpha=0;
		break;
	  case 3:
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
		gAlpha=64;
		break;
	}*/
	//syslog(77," CMD: P4TF: %d,%d TP=%d",textAddrX,textAddrY,textTP);
	//syslog(77," CMD: P4TG: %d,%d,%d,%d",baseAddr[3],baseAddr[15],baseAddr[39],baseAddr[27]);
	//syslog(77," CMD: P4TF: %d,%d %02x",lx0,ly0,baseAddr[3]);

    clutP  = (gpuData[2]>>12) & 0x7fff0;//0xffff0;
	setupTexture(clutP);
	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		if(baseAddr[3]&1){
			glColor4ub(255,255,255,gAlpha);
		}else{
			glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],gAlpha);
		}
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		if(baseAddr[3]&1){
			glColor3ub(255,255,255);
		}else{
			glColor3ub(baseAddr[0],baseAddr[1],baseAddr[2]);
		}
	}
	glBegin(GL_POLYGON);
		glTexCoord2s(tx0,ty0);
		glVertex2s(lx0,ly0+dispOffset);
		glTexCoord2s(tx1,ty1);
		glVertex2s(lx1,ly1+dispOffset);
		glTexCoord2s(tx3,ty3);
		glVertex2s(lx3,ly3+dispOffset);
		glTexCoord2s(tx2,ty2);
		glVertex2s(lx2,ly2+dispOffset);
	glEnd();
	glBindTexture(GL_TEXTURE_2D,nullid);
}

void primPolyGT3(unsigned char *baseAddr)
{	
/* 
	Goraud Shaded Textured Polygon 3 vertices
*/
	unsigned long *gpuData = (unsigned long *)baseAddr;
	short *gpuPoint=((short *) baseAddr);
	long clutP;
	short lx0,ly0,lx1,ly1,lx2,ly2;
	short tx0,ty0,tx1,ty1,tx2,ty2;
	unsigned short gpuDataX;
	unsigned char alpha;

	lx0 = gpuPoint[2]+ofsX;
	ly0 = gpuPoint[3]+ofsY;
	lx1 = gpuPoint[8]+ofsX;
	ly1 = gpuPoint[9]+ofsY;
	lx2 = gpuPoint[14]+ofsX;
	ly2 = gpuPoint[15]+ofsY;

	tx0 = (short)(gpuData[2] & 0xff);
	ty0 = (short)((gpuData[2]>>8) & 0xff);
	tx1 = (short)(gpuData[5] & 0xff);
	ty1 = (short)((gpuData[5]>>8) & 0xff);
	tx2 = (short)(gpuData[8] & 0xff);
	ty2 = (short)((gpuData[8]>>8) & 0xff);
	//syslog(77,"!GPU! P3TG %d,%d %d,%d %d,%d\n",lx0,ly0,lx1,ly1,lx2,ly2);

	gpuDataX=(unsigned short)(gpuData[5]>>16);
	textAddrX = (gpuDataX&0xf)<<6;
	textAddrX2 = (gpuDataX&0xf)<<7;
	textAddrY = (((gpuDataX)<<4)&0x100);//+(((gpuDataX)>>2)&0x200);
	textTP = (gpuDataX & 0x180) >> 7;
	long tempABR;
	tempABR = (gpuDataX >> 5) & 0x3;
	if(textABR!=tempABR){
		textABR=tempABR;
		glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
		gAlpha=TransRate[textABR].alpha;
		transparent=TRUE;
	}
	/*switch(textABR){
	  case 0:
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		gAlpha=128;
		break;
	  case 1:
		glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA);
		gAlpha=0;
		break;
	  case 2:
		glBlendFunc(GL_ZERO,GL_ONE_MINUS_SRC_COLOR);
		gAlpha=0;
		break;
	  case 3:
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
		gAlpha=64;
		break;
	}*/

	//syslog(77," CMD: P3TG: %d,%d TP=%d",textAddrX,textAddrY,textTP);
	//syslog(77," CMD: P3TG: %d,%d %02x",lx0,ly0,baseAddr[3]);

    clutP  = (gpuData[2]>>12) & 0x7fff0;//0xffff0;
	setupTexture(clutP);
	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		alpha=gAlpha;
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		alpha=255;
	}
	glBegin(GL_POLYGON);
		glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],alpha);
		glTexCoord2s(tx0,ty0);
		glVertex2s(lx0,ly0+dispOffset);
		glColor4ub(baseAddr[12],baseAddr[13],baseAddr[14],alpha);
		glTexCoord2s(tx1,ty1);
		glVertex2s(lx1,ly1+dispOffset);
		glColor4ub(baseAddr[24],baseAddr[25],baseAddr[26],alpha);
		glTexCoord2s(tx2,ty2);
		glVertex2s(lx2,ly2+dispOffset);
	glEnd();
	glBindTexture(GL_TEXTURE_2D,nullid);
}

void primPolyGT4(unsigned char *baseAddr)
{	
/* 
	Goraud Shaded Textured Polygon 4 vertices
*/
	unsigned long *gpuData = (unsigned long *)baseAddr;
	short *gpuPoint=((short *) baseAddr);
	long clutP;
	short lx0,ly0,lx1,ly1,lx2,ly2,lx3,ly3;
	short tx0,ty0,tx1,ty1,tx2,ty2,tx3,ty3;
	unsigned short gpuDataX;
	unsigned char alpha;

	lx0 = gpuPoint[2]+ofsX;
	ly0 = gpuPoint[3]+ofsY;
	lx1 = gpuPoint[8]+ofsX;
	ly1 = gpuPoint[9]+ofsY;
	lx2 = gpuPoint[14]+ofsX;
	ly2 = gpuPoint[15]+ofsY;
	lx3 = gpuPoint[20]+ofsX;
	ly3 = gpuPoint[21]+ofsY;

	tx0 = (short)(gpuData[2] & 0xff);
	ty0 = (short)((gpuData[2]>>8) & 0xff);
	tx1 = (short)(gpuData[5] & 0xff);
	ty1 = (short)((gpuData[5]>>8) & 0xff);
	tx2 = (short)(gpuData[8] & 0xff);
	ty2 = (short)((gpuData[8]>>8) & 0xff);
	tx3 = (short)(gpuData[11] & 0xff);
	ty3 = (short)((gpuData[11]>>8) & 0xff);

	gpuDataX=(unsigned short)(gpuData[5]>>16);
	textAddrX = (gpuDataX&0xf)<<6;
	textAddrX2 = (gpuDataX&0xf)<<7;
	textAddrY = (((gpuDataX)<<4)&0x100);//+(((gpuDataX)>>2)&0x200);
	textTP = (gpuDataX & 0x180) >> 7;
	long tempABR;
	tempABR = (gpuDataX >> 5) & 0x3;
	if(textABR!=tempABR){
		textABR=tempABR;
		glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
		gAlpha=TransRate[textABR].alpha;
		transparent=TRUE;
	}
	/*switch(textABR){
	  case 0:
		glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
		gAlpha=128;
		break;
	  case 1:
		glBlendFunc(GL_ONE,GL_ONE_MINUS_SRC_ALPHA);
		gAlpha=0;
		break;
	  case 2:
		glBlendFunc(GL_ZERO,GL_ONE_MINUS_SRC_COLOR);
		gAlpha=0;
		break;
	  case 3:
		glBlendFunc(GL_SRC_ALPHA,GL_ONE);
		gAlpha=64;
		break;
	}*/

	//syslog(77," CMD: P4TG: %d,%d TP=%d",textAddrX,textAddrY,textTP);
	//syslog(77," CMD: P4TG: %d,%d %02x",lx0,ly0,baseAddr[3]);

    clutP  = (gpuData[2]>>12) & 0x7fff0;//0xffff0;
	setupTexture(clutP);
	if(baseAddr[3]&2){
		if(TransRate[textABR].change&&!transparent){
			glBlendFunc(TransRate[textABR].src,TransRate[textABR].dst);
			transparent=TRUE;
		}
		alpha=gAlpha;
	}else{
		if(TransRate[textABR].change&&transparent){
			glBlendFunc(TransRate[0].src,TransRate[0].dst);
			transparent=FALSE;
		}
		alpha=255;
	}
	glBegin(GL_POLYGON);
		glColor4ub(baseAddr[0],baseAddr[1],baseAddr[2],alpha);
		glTexCoord2s(tx0,ty0);
		glVertex2s(lx0,ly0+dispOffset);
		glColor4ub(baseAddr[12],baseAddr[13],baseAddr[14],alpha);
		glTexCoord2s(tx1,ty1);
		glVertex2s(lx1,ly1+dispOffset);
		glColor4ub(baseAddr[36],baseAddr[37],baseAddr[38],alpha);
		glTexCoord2s(tx3,ty3);
		glVertex2s(lx3,ly3+dispOffset);
		glColor4ub(baseAddr[24],baseAddr[25],baseAddr[26],alpha);
		glTexCoord2s(tx2,ty2);
		glVertex2s(lx2,ly2+dispOffset);
	glEnd();
	glBindTexture(GL_TEXTURE_2D,nullid);
}


void primNI(unsigned char *bA)
{
	// primitive not implemented
	if (bA[3]) syslog(255,"PRIM: *NI* %02x",bA[3]);
}


unsigned char primTableC[256] =
{
	// 00
	0,0,3,0,0,0,0,0,
	// 08
	0,0,0,0,0,0,0,0,
	// 10
	0,0,0,0,0,0,0,0,
	// 18
	0,0,0,0,0,0,0,0,
	// 20
	4,4,4,4,7,7,7,7,
	// 28
	5,5,5,5,9,9,9,9,
	// 30
	6,6,6,6,9,9,9,9,
	// 38
	8,8,8,8,12,12,12,12,
	// 40
	3,3,3,3,0,0,0,0,    //  LineF2
	// 48
	5,5,5,5,6,6,6,6,	//	LineF3	LineF4
	// 50
	4,4,4,4,0,0,0,0,    //  LineG2
	// 58
	7,7,7,7,9,9,9,9,	//	LineG3	LineG4
	// 60
	3,3,3,3,4,4,4,4,	//	Tile	Sprt
	// 68
	2,2,2,2,0,0,0,0,	//	Tile1
	// 70
	2,2,2,2,3,3,3,3,	//	Tile8	Sprt8
	// 78
	2,2,2,2,3,3,3,3,	//	Tile16	Sprt16
	// 80
	4,0,0,0,0,0,0,0,
	// 88
	0,0,0,0,0,0,0,0,
	// 90
	0,0,0,0,0,0,0,0,
	// 98
	0,0,0,0,0,0,0,0,
	// a0
	3,0,0,0,0,0,0,0,
	// a8
	0,0,0,0,0,0,0,0,
	// b0
	0,0,0,0,0,0,0,0,
	// b8
	0,0,0,0,0,0,0,0,
	// c0
	3,0,0,0,0,0,0,0,
	// c8
	0,0,0,0,0,0,0,0,
	// d0
	0,0,0,0,0,0,0,0,
	// d8
	0,0,0,0,0,0,0,0,
	// e0
	0,1,1,1,1,1,1,0,
	// e8
	0,0,0,0,0,0,0,0,
	// f0
	0,0,0,0,0,0,0,0,
	// f8
	0,0,0,0,0,0,0,0

};

void (*primTableJ[256])(unsigned char *) = 
{
	// 00
	primNI,primNI,primBlkFill,primNI,primNI,primNI,primNI,primNI,
	// 08
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// 10
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// 18
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// 20
	primPolyF3,primPolyF3,primPolyF3,primPolyF3,primPolyFT3,primPolyFT3,primPolyFT3,primPolyFT3,
	// 28
	primPolyF4,primPolyF4,primPolyF4,primPolyF4,primPolyFT4,primPolyFT4,primPolyFT4,primPolyFT4,
	// 30
	primPolyG3,primPolyG3,primPolyG3,primPolyG3,primPolyGT3,primPolyGT3,primPolyGT3,primPolyGT3,
	// 38
	primPolyG4,primPolyG4,primPolyG4,primPolyG4,primPolyGT4,primPolyGT4,primPolyGT4,primPolyGT4,
	// 40
	primLineF2,primLineF2,primLineF2,primLineF2,primNI,primNI,primNI,primNI,
	// 48
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// 50
	primLineG2,primLineG2,primLineG2,primLineG2,primNI,primNI,primNI,primNI,
	// 58
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// 60
	primTileS,primNI,primNI,primNI,primSprtS,primSprtS,primSprtS,primSprtS,
	// 68
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// 70
	primNI,primNI,primNI,primNI,primSprt8,primSprt8,primSprt8,primSprt8,
	// 78
	primNI,primNI,primNI,primNI,primSprt16,primSprt16,primSprt16,primSprt16,
	// 80
	primMoveImage,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// 88
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// 90
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// 98
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// a0
	primLoadImage,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// a8
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// b0
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// b8
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// c0
	primStoreImage,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// c8
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// d0
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// d8
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// e0
	primNI,cmdTexturePage,cmdTextureWindow,cmdDrawAreaStart,cmdDrawAreaEnd,cmdDrawOffset,cmdSTP,primNI,
	// e8
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// f0
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI,
	// f8
	primNI,primNI,primNI,primNI,primNI,primNI,primNI,primNI
};

